/*
 * pictoppm.c -- convert .pic files to ppm (Mouse systems PC paint or grasp
 * .gl files)
 * 
 * Author:  George Phillips <phillips@cs.ubc.ca>
 * 
 * Extended to handle sequences of .pic and .clp files by
 * Burkhard Neidecker-Lutz <neideck@nestvx.enet.dec.com>
 */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

int blocksleft;
int curlen = 0;
int spec;
int bytesleft;
int verbose = 1;
unsigned char cmap[256 * 3];
int gotcolormap  = 0;	

void ueof()
{
  fprintf(stderr, "unexpected EOF\n");
  exit(1);
}


int rd(f,b,len)
FILE *f;
unsigned char *b;
int len;
{
  int             i;
  int             c;

  for (i = 0; i < len; i++) {
    if ((c = getc(f)) == EOF) {
      ueof();
    }
    b[i] = (c & 255);
  }
  return 1;
}

int brd(f,b,len)
FILE *f;
unsigned char *b;
int len;
{
  unsigned char   buf[5];

  if (curlen <= 0) {
    if (blocksleft == 0) {
      fprintf(stderr, "out of blocks (need %d more bytes)\n", bytesleft);
      exit(1);
    }
    rd(f,buf, 5);

    curlen = buf[0] + buf[1] * 256 - 5;
    /* size of buffer needed to store next block of pixels? */
    /* bufsize = buf[2] + buf[3] * 256; */
    spec = buf[4];
    blocksleft--;
  }
  /* this should be done right and allow for overlapping reads! :-( */

  if (len > curlen) {
    fprintf(stderr, "arg, need to read over block boundry!\n");
    exit(1);
  }
  rd(f,b,len);

  curlen -= len;

  return 1;
}

void convertfile(name)
char *name;
{
  unsigned char   buf[32];
  int             hascmap;
  int             i, r, c;
  int             width, height, bpp, w, h;
  char	*n;
  unsigned char  *img;
  unsigned char  *p;
  char ofname[100];
  FILE *inf,*of;

  if (!(inf = fopen(name,"r"))) {	
    perror(name);
    exit(1);
  }
  
  for (i = 0, n = name; *n != '.'; i++, n++) {	/* Find basename of file */
    ofname[i] = *n;
  }
  ofname[i] = '.';		/* Too lazy to use a string function here */
  ofname[++i] = 'p';
  ofname[++i] = 'p';
  ofname[++i] = 'm';
  ofname[++i] = 0;

  if (!(of = fopen(ofname,"w"))) {	/* Open output file */
    perror(ofname);
    exit(1);
  }
  
  rd(inf,buf, 17);

  if (buf[15] != 0) {
    rd(inf,buf + 17, 2);
  }
  if (buf[0] != 0x34 || buf[1] != 0x12) {
    fprintf(stderr, "not a .pic file\n");
    exit(1);
  }
  if (buf[12] != 'L' && buf[12] != 'C')
    fprintf(stderr, "I may not do the right thing...\n");

  hascmap = buf[13] & 4;

  width = buf[2] + buf[3] * 256;
  height = buf[4] + buf[5] * 256;
  bpp = buf[10];
  /* I'm not real sure about the bits-per-pixel, above */

  if (verbose)
    fprintf(stderr, "%d x %d x %d type %c, %s colourmap\n",
	    width, height, 1 << bpp, buf[12], hascmap ? "has a" : "no");


  /* allocation equation depends on 8 | width (| = divides) */
  if ((p = img =
	 (unsigned char *) malloc(bytesleft = width * height * bpp / 8)) == NULL) {
    fprintf(stderr, "not enough memory\n");
    exit(1);
  }
  if (hascmap) {
    rd(inf,cmap, 256 * 3);

    for (i = 0; i < 256 * 3; i++) {
      cmap[i] *= 4;
    }
    gotcolormap = 1;
  } else if (!gotcolormap) {
    for (i = 0; i < 256 * 3; i++)
      cmap[i] = i / 3;
  }

  rd(inf,buf, 2);

  blocksleft = buf[0] + buf[1] * 256;

  while (brd(inf,buf, 1)) {
    c = buf[0];
    if (c == spec) {
      brd(inf,buf, 1);
      r = buf[0];
      if (r == 0) {
	brd(inf,buf, 3);
	r = buf[1] * 256 + buf[0];
	c = buf[2];
      } else {
	brd(inf,buf, 1);
	c = buf[0];
      }
    } else {
      r = 1;
    }
    while (r-- > 0) {
      if (--bytesleft < 0) {
	fprintf(stderr, "file too long, ignoring extra\n");
	goto dump;
      }
      *p++ = c;
    }
    if (bytesleft == 0)
      break;
  }

dump:

  if (bpp == 8) {
    fprintf(of,"P6\n%d %d\n255\n", width, height);

    for (h = height - 1; h >= 0; h--)
      for (w = 0; w < width; w++) {
	c = img[h * width + w];
	putc(cmap[c * 3],of);
	putc(cmap[c * 3 + 1],of);
	putc(cmap[c * 3 + 2],of);
      }
  } else {
    fprintf(of,"P4\n%d %d\n", width, height);

    for (h = height - 1; h >= 0; h--)
      for (w = 0; w < width / 8; w++)
	putc(255 - img[h * width / 8 + w],of);
  }
  free(p);
  fclose(inf);
  fclose(of);
}

int main(argc, argv)
int  argc;
char *argv[];
{
  int i;

  for (i = 1; i < argc ; i++) {
    convertfile(argv[i]);
  }
  return(0);
}

